MODULE fsupport
  USE, INTRINSIC:: iso_c_binding
#ifdef _OPENACC
  USE openacc
#endif
  IMPLICIT NONE
  PRIVATE
  PUBLIC :: dp
  PUBLIC :: t_domain
  PUBLIC :: get_domain
  PUBLIC :: init_field
  PUBLIC :: zero_field
  PUBLIC :: physop
  PUBLIC :: f2c_default_run_all
  !PUBLIC :: show_field
  PUBLIC :: f2c_kokkos_initialize
  PUBLIC :: f2c_kokkos_finalize
  INTEGER, PARAMETER :: dp = c_double

  TYPE, BIND(c):: t_domain
    INTEGER(c_int) :: ncells
    INTEGER(c_int) :: nlev
    INTEGER(c_int) :: nblocks
    INTEGER(c_int) :: nproma
  END TYPE t_domain

  INTERFACE
    SUBROUTINE f2c_kokkos_initialize() BIND(c)
    END SUBROUTINE f2c_kokkos_initialize

    SUBROUTINE f2c_kokkos_finalize() BIND(c)
    END SUBROUTINE f2c_kokkos_finalize

    SUBROUTINE f2c_get_domain(dom) BIND(c)
      IMPORT t_domain
      TYPE(t_domain) :: dom
    END SUBROUTINE f2c_get_domain

    SUBROUTINE f2c_default_run_all(field, ref_in, ref_out, nblocks, nlev, nproma) BIND(c)
      IMPORT c_double
      REAL(c_double) :: field(*), ref_in(*), ref_out(*)
      INTEGER, VALUE ::  nblocks, nlev, nproma
    END SUBROUTINE f2c_default_run_all
  END INTERFACE

CONTAINS

  SUBROUTINE get_domain(dom)
    TYPE(t_domain), INTENT(out) :: dom
    CALL f2c_get_domain(dom)
  END SUBROUTINE get_domain

  SUBROUTINE init_field(field)
    REAL(dp), INTENT(out) :: field(:,:,:)
    INTEGER :: j,k,i
    INTEGER :: nj,nk,ni
    nj = SIZE(field,3)
    nk = SIZE(field,2)
    ni = SIZE(field,1)
    !PRINT*,'ni,nk,nj=',ni,nk,nj
    DO j = 1, nj
!$ACC parallel default(present)
!$ACC LOOP GANG VECTOR COLLAPSE(2)
      DO k = 1, nk
        DO i = 1, ni
          field(i,k,j) = REAL(i-1 + (k-1)*ni + (j-1)*ni*nk, dp)
          !PRINT*,'i,k,j,val=',i,k,j, field(i,k,j)
        ENDDO
      ENDDO
!$ACC END PARALLEL
    ENDDO
  END SUBROUTINE init_field

  SUBROUTINE zero_field(v)
    REAL(dp), INTENT(out) :: v(:,:,:)
    INTEGER :: j,k,i
    INTEGER :: nj,nk,ni
    nj = SIZE(v,3)
    nk = SIZE(v,2)
    ni = SIZE(v,1)
    DO j = 1, nj
!$ACC parallel default(present)
!$ACC LOOP GANG VECTOR COLLAPSE(2)
      DO k = 1, nk
        DO i = 1, ni
          v(i,k,j) = 0.0_dp
        ENDDO
      ENDDO
!$ACC END PARALLEL
    ENDDO
  END SUBROUTINE zero_field

  SUBROUTINE physop(v)
    REAL(dp), INTENT(out) :: v(:,:,:)
    INTEGER :: j,k,i
    INTEGER :: nj,nk,ni
    nj = SIZE(v,3)
    nk = SIZE(v,2)
    ni = SIZE(v,1)
    PRINT*,'physop: ni,nk,nj=',ni,nk,nj
    DO j = 1, nj
!$ACC PARALLEL DEFAULT(PRESENT)
      k = 1
!$ACC LOOP GANG VECTOR
      DO i = 1, ni
        v(i,k,j) = v(i,k,j) + v(i,k+1,j)/10000
      ENDDO
!$ACC LOOP SEQ
      DO k = 2, nk-1
!$ACC LOOP GANG VECTOR
        DO i = 1, ni
          v(i,k,j) = v(i,k,j) +  v(i,k-1,j)/100 + v(i,k+1,j)/10000;
        ENDDO
      ENDDO
      k = nk
!$ACC LOOP GANG VECTOR
      DO i = 1, ni
        v(i,k,j) = v(i,k,j) + v(i,k-1,j)/100;
      ENDDO
!$ACC END PARALLEL
    ENDDO
  END SUBROUTINE physop

#if 0
  SUBROUTINE show_field(label,f)
    CHARACTER(len=*), INTENT(in) :: label
    REAL(dp), INTENT(in) :: f(:,:,:)
    INTEGER :: i,k,j
    INTEGER :: ni,nk,nj
    nj = SIZE(f,1)
    nk = SIZE(f,2)
    ni = SIZE(f,3)
    DO j = 1, nj
!$ACC PARALLEL DEFAULT(PRESENT)
!$ACC LOOP SEQ
      DO k = 1, nk
        DO i = 1, ni
          PRINT '(A,3I8,F)','show_field: '//label,j,k,i,f(i,k,j)
        ENDDO
      ENDDO
!$ACC END PARALLEL
    ENDDO
  END SUBROUTINE show_field
#endif
END MODULE fsupport

PROGRAM fdriver
  USE fsupport, ONLY: dp, t_domain, get_domain, init_field, zero_field, physop, &
       & f2c_default_run_all, f2c_kokkos_initialize, f2c_kokkos_finalize
  IMPLICIT NONE

  TYPE(t_domain) :: dom
  REAL(dp), ALLOCATABLE :: field(:,:,:), ref_in(:,:,:), ref_out(:,:,:)
  CALL f2c_kokkos_initialize()
  PRINT*,'fdriver::start'
  CALL get_domain(dom)
  PRINT '(A,4I8)','ncells,lev,nblocks,nproma=',dom%ncells,dom%nlev,dom%nblocks,dom%nproma

  ALLOCATE( field(dom%nproma,dom%nlev,dom%nblocks), &
       &    ref_in(dom%nproma,dom%nlev,dom%nblocks), &
       &    ref_out(dom%nproma,dom%nlev,dom%nblocks) )

!$acc enter data create(field, ref_in, ref_out)
  CALL zero_field(field)
  CALL init_field(ref_in)
  CALL init_field(ref_out)
  CALL physop(ref_out)

!$acc host_data use_device(field, ref_in, ref_out)
  CALL f2c_default_run_all(field, ref_in, ref_out, dom%nblocks, dom%nlev,  dom%nproma)
!$acc end host_data

!$acc exit data copyout(field, ref_in, ref_out)

  CALL f2c_kokkos_finalize()
  PRINT*,'fdriver:end'
END PROGRAM fdriver
