#ifndef ICON_LOOPS
#define ICON_LOOPS

#include <Kokkos_Core.hpp>

#define outer_lambda [=]
#define inner_lambda KOKKOS_LAMBDA

namespace icon_loops {

constexpr bool exec_on_device() {
  using ExecSpace = Kokkos::DefaultExecutionSpace;
#ifdef KOKKOS_ENABLE_HIP
  constexpr bool exec_on_hip = std::is_same<ExecSpace,Kokkos::HIP>::value;
#else
  constexpr bool exec_on_hip = false;
#endif
#ifdef KOKKOS_ENABLE_CUDA
  constexpr bool exec_on_cuda = std::is_same<ExecSpace,Kokkos::Cuda>::value;
#else
  constexpr bool exec_on_cuda = false;
#endif
  return exec_on_cuda || exec_on_hip;
}

template<class FunctorType>
inline void outer_loop(const std::string& str, const int start, const int end, const FunctorType& functor) {
  if constexpr (exec_on_device()) {
    for(int i=start; i<end; ++i) {
      functor(i);
    }
  } else {
    Kokkos::parallel_for(str,  Kokkos::RangePolicy<>(start, end), functor);
  }
}

template<class FunctorType>
inline void outer_loop(const int start, const int end, const FunctorType& functor) {
  if constexpr (exec_on_device()) {
    for(int i=start; i<end; ++i) {
      functor(i);
    }
  } else {
    Kokkos::parallel_for("",  Kokkos::RangePolicy<>(start, end), functor);
  }
}

template<class FunctorType>
inline void outer_loop(const int end, const FunctorType& functor) {
  if constexpr (exec_on_device()) {
    for(int i=0; i<end; ++i) {
      functor(i);
    }
  } else {
    Kokkos::parallel_for("",  Kokkos::RangePolicy<>(0, end), functor);
  }
}

template<class FunctorType>
inline void inner_loop(const std::string& str, const int start, const int end, const FunctorType& functor) {
  if constexpr (exec_on_device()) {
    Kokkos::parallel_for(str,Kokkos::RangePolicy<>(start, end), functor);
  } else {
    for(int i=start; i<end; ++i) {
      functor(i);
    }
  }
}

template<class FunctorType>
inline void inner_loop(const int start, const int end, const FunctorType& functor) {
  if constexpr (exec_on_device()) {
    Kokkos::parallel_for("",Kokkos::RangePolicy<>(start, end), functor);
  } else {
    for(int i=start; i<end; ++i) {
      functor(i);
    }
  }
}

template<class FunctorType>
inline void inner_loop(const int end, const FunctorType& functor) {
  if constexpr (exec_on_device()) {
    Kokkos::parallel_for("",Kokkos::RangePolicy<>(0, end), functor);
  } else {
    for(int i=0; i<end; ++i) {
      functor(i);
    }
  }
}

} // namespace icon_loops
#endif
