#include <Kokkos_Core.hpp>
#include <functional>
#undef NDEBUG // always use assertion
#include <cassert>
#include <iostream>
#include <utility>
#include <cstring>

#include "Kokkos_Timer.hpp"
#include "icon_loops.hpp"
using icon_loops::outer_loop, icon_loops::inner_loop, icon_loops::exec_on_device ;

using space_t = Kokkos::DefaultExecutionSpace::memory_space;
typedef Kokkos::MDRangePolicy<Kokkos::DefaultExecutionSpace, Kokkos::IndexType<int>, Kokkos::Rank<3>> md_range_policy;

Kokkos::Timer timer;

#define ENABLE_CHECK_BOUNDS

static int verbose_level = 0;

struct AllLevels {
  constexpr static int value = 0;
};
constexpr AllLevels all_levels;

#if defined(DEMO_DEVICE)
# define HOST_DEVICE_ATTRIBUTES __host__ __device__
#else
# define HOST_DEVICE_ATTRIBUTES
#endif

//macro approach:
#ifdef DEMO_DEVICE
#  define outer_for(STR, START, END, LAMBDA) for(int jb=START; jb<END; ++jb) {LAMBDA(jb);};
#  define inner_for(STR, START, END, LAMBDA) Kokkos::parallel_for(STR,  Kokkos::RangePolicy<>(START, END), LAMBDA)
#else
#  define outer_for(STR, START, END, LAMBDA) Kokkos::parallel_for(STR,  Kokkos::RangePolicy<>(START, END), LAMBDA)
#  define inner_for(STR, START, END, LAMBDA) for(int jc=START; jc<END; ++jc) {LAMBDA(jc);};
#endif

static void validate(double* array, const double *ref, int nblocks, int nlev, int nproma) {
  for (int i = 0; i < nblocks * nlev * nproma; ++i) {
    //printf("check i=%d, array=%f, ref=%f\n",i, array[i] , ref[i]);
    assert(array[i] == ref[i]);
  }
}


inline static HOST_DEVICE_ATTRIBUTES double  ferror(double x1, double x0) {
  constexpr double eps = 1.e-14;
  if (fabs(x0) > eps) {
    return fabs(x1-x0)/fabs(x0);
  } else {
    return fabs(x1-x0);
  }
}

static void validate_max_error(double* array, const double *ref, int nblocks, int nlev, int nproma) {
  const int ntot = nblocks*nlev*nproma;
  double totmax;
  Kokkos::parallel_reduce("validator", ntot, KOKKOS_LAMBDA(const int &i, double &emax) {
      double err = ferror(array[i],ref[i]);
      emax = fmax(emax, err);
    }, Kokkos::Max<double>(totmax));
  if (verbose_level > 0) printf("validate_max_error: total max error=%.10e\n",totmax);
  assert(totmax < 1.e-14);
}

KOKKOS_INLINE_FUNCTION void check_bounds(int i0, int i1, int i2, int n0, int n1, int n2) {
#ifdef ENABLE_CHECK_BOUNDS
  assert(i0 >= 0 && i1 >= 0 && i2 >= 0 && i0 < n0 && i1 < n1 && i2 < n2);
#endif
}

KOKKOS_INLINE_FUNCTION void check_bounds(int i0, int i1, int n0, int n1) {
#ifdef ENABLE_CHECK_BOUNDS
  assert(i0 >= 0 && i1 >= 0 && i0 < n0 && i1 < n1 );
#endif
}


template<typename ViewType>
inline HOST_DEVICE_ATTRIBUTES void update_column(ViewType &v) {
  const int nlev = v.extent(0);

  {int k = 0; v(k) += v(k+1)/10000;}
  for (int k=1; k < nlev-1; ++k) {
    v(k) += v(k-1)/100 + v(k+1)/10000;
  }
  {int k = nlev-1; v(k) += v(k-1)/100;}
}


template<typename ViewType>
inline HOST_DEVICE_ATTRIBUTES void update_field(ViewType &v, int i0, AllLevels all, int i2) {
  const int nlev = v.extent(1);
  check_bounds(i0, i2, v.extent(0), v.extent(2));
  auto column = subview(v,i0,Kokkos::ALL, i2); // why can't we use the RHS as argument to update_column directly?
  update_column(column);
}

void scenario_1(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 1: Default layout; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nblocks, nlev, nproma);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          update_field(d_view, jb, all_levels, jc);
        });
  Kokkos::fence();

  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);

}

void scenario_1_noprint(double* array, int nblocks, int nlev, int nproma) {
  printf("scenario_1_noprint\n");
  scenario_1(array, nblocks, nlev, nproma, false);
}


void scenario_1b(double* array, int nblocks, int nlev, int nproma) {
  std::cout << "scenario 1b: always LayoutRight; view(array, nblocks, nlev, nproma); parallel: nproma ----- "
            << std::endl;
  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);

  timer.reset();

  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          update_field(d_view, jb, all_levels, jc);
        });
  Kokkos::fence();

  printf("Time = %f ms\n\n", timer.seconds() * 1000);

}


void scenario_1c(double* array, int nblocks, int nlev, int nproma) {
  std::cout << "scenario 1c: Default layout; view(array, ncells, nlev); d_view(jc, jk) ----- "
              << std::endl;
  int ncells = nblocks*nproma;

  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> d_view3d(array, nproma, nlev, nblocks);
  //Kokkos::View<double**, Kokkos::MemoryUnmanaged> d_view(array, ncells, nlev);
  Kokkos::View<double**> d_view2d("d_view2d", ncells, nlev);
  Kokkos::Timer tr_timer;
  Kokkos::Timer total_timer;
  total_timer.reset();

  // view3d ->  view2d:
  tr_timer.reset();
  outer_loop(nblocks, outer_lambda(const int jb){
      inner_loop(nproma, inner_lambda(const int jc) {
          int jh = jb*nproma + jc;
          for (int jk = 0; jk < nlev; ++jk) {
            d_view2d(jh, jk) =  d_view3d(jc,jk,jb);
          }
        });
    });
  Kokkos::fence();
  printf("Transposition 3d -> 2d = %f ms\n", tr_timer.seconds() * 1000);
  
  timer.reset();
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, ncells), KOKKOS_LAMBDA(const int jc) {
          check_bounds(jc, nlev-2, d_view2d.extent(0), d_view2d.extent(1));
           auto column = subview(d_view2d,jc,Kokkos::ALL);
           update_column(column);
        });
  Kokkos::fence();
  printf("Kernel = %f ms\n", timer.seconds() * 1000);

  // view2d ->  view3d:
  tr_timer.reset();
  outer_loop(nblocks, outer_lambda(const int jb){
      inner_loop(nproma, inner_lambda(const int jc) {
          int jh = jb*nproma + jc;
          for (int jk = 0; jk < nlev; ++jk) {
            d_view3d(jc,jk,jb) = d_view2d(jh, jk);
          }
        });
    });
  Kokkos::fence();
  printf("Transposition 2d -> 3d = %f ms\n", tr_timer.seconds() * 1000);

  printf("Time = %f ms\n\n", total_timer.seconds() * 1000);
}


void scenario_1cmacro(double* array, int nblocks, int nlev, int nproma) {
    std::cout << "scenario 1cmacro: +ACC CPU:LayoutLeft GPU:LayoutLeft; as ICON pointer view(array, ncells, nlev); d_view(jc, jk) ----- " << std::endl;

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  int ncells = nblocks*nproma;

#if defined(DEMO_DEVICE)
  Kokkos::View<double**, Kokkos::LayoutLeft> d_view("view2d", ncells, nlev);
#else
  Kokkos::View<double**, Kokkos::LayoutLeft> d_view("view2d", ncells, nlev);
#endif

  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> d_view3d(array, nproma, nlev, nblocks);
  Kokkos::Timer tr_timer;
  Kokkos::Timer total_timer;
  total_timer.reset();

  // view3d ->  view2d:
  tr_timer.reset();
  outer_loop(nblocks, outer_lambda(const int jb){
      inner_loop(nproma, inner_lambda(const int jc) {
          int jh = jb*nproma + jc;
          for (int jk = 0; jk < nlev; ++jk) {
            d_view(jh, jk) =  d_view3d(jc,jk,jb);
          }
        });
    });
  Kokkos::fence();
  printf("Transposition 3d -> 2d = %f ms\n", tr_timer.seconds() * 1000);

  timer.reset();

    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, ncells), KOKKOS_LAMBDA(const int jc) {
          check_bounds(jc, nlev-2, d_view.extent(0), d_view.extent(1));
           auto column = subview(d_view,jc,Kokkos::ALL);
           update_column(column);
        });
  Kokkos::fence();

  printf("Kernel = %f ms\n", timer.seconds() * 1000);

  // view2d ->  view3d:
  tr_timer.reset();
  outer_loop(nblocks, outer_lambda(const int jb){
      inner_loop(nproma, inner_lambda(const int jc) {
          int jh = jb*nproma + jc;
          for (int jk = 0; jk < nlev; ++jk) {
            d_view3d(jc,jk,jb) = d_view(jh, jk);
          }
        });
    });
  Kokkos::fence();
  printf("Transposition 2d -> 3d = %f ms\n", tr_timer.seconds() * 1000);
  printf("Time = %f ms\n\n", total_timer.seconds() * 1000);

}

#if 0
void scenario_2(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 2: Right layout; view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- " << std::endl;

  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev,
                                                                                                nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();

  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jc * nlev * nblocks + jk * nblocks + jb;
            check_bounds(jc, jk, jb, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  //validate(array, nblocks, nlev, nproma);
}
#endif

#if 0
void scenario_2b(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 2b: Right 2b layout; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nblocks, nlev,
                                                                                                nproma);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();

  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            check_bounds(jb, jk, jc, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jb, jk, jc) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}
#endif

#if 0
// slow on CPU
void scenario_3(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 3: Left layout; view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- " << std::endl;

  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev,
                                                                                               nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            check_bounds(jc, jk, jb, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}
#endif

#if 0
void scenario_4(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 4: Default layout; view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p = jc * nlev * nblocks + jk * nblocks + jb;
            check_bounds(jc, jk, jb, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });
  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}
#endif

#if 0
void scenario_4b(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print)
    std::cout << "scenario 4b (Dmitry's solution): view(array, nproma, nlev, nblocks); d_view(jc, jk, jb) ----- "
              << std::endl;

  Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t   = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view_tmp = Kokkos::create_mirror_view_and_copy(space_t(), view);
  Kokkos::View<double***, space_t> d_view("d_view", nproma, nlev, nblocks);
  Kokkos::deep_copy(d_view_tmp, view);
  Kokkos::deep_copy(d_view, d_view_tmp);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb)
    Kokkos::parallel_for(
        "", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          for (int jk = 0; jk < nlev; ++jk) {
            int p              = jc * nlev * nblocks + jk * nblocks + jb;
            d_view(jc, jk, jb) = p;

            //  printf("%f ", d_view(jb, jk, jc));
          }
        });

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(d_view_tmp, d_view);
  Kokkos::deep_copy(view, d_view_tmp);
  validate(array, nblocks, nlev, nproma);
}
#endif

#if defined(DEMO_DEVICE)
  using Layout = Kokkos::LayoutLeft;
#else
  using Layout = Kokkos::LayoutRight;
#endif

template <class T> struct LoopFunctor {
  Kokkos::View<T**, Kokkos::LayoutStride> view;
  int nproma, nlev, nblocks, jb;

  LoopFunctor(int nproma, int nlev, int nblocks, int jb, Kokkos::View<T***, Kokkos::LayoutStride> d_view)
      : nproma(nproma), nlev(nlev), nblocks(nblocks), jb(jb) {
    view = Kokkos::subview(d_view, Kokkos::ALL, Kokkos::ALL, jb);
  }

  KOKKOS_INLINE_FUNCTION
  void operator()(const int jc) const {
    for (int jk = 0; jk < nlev; ++jk) {
#if defined(DEMO_DEVICE)
      int p = jb * nlev * nproma + jk * nproma + jc;
#else
      int p = jc * nlev * nblocks + jk * nblocks + jb;
#endif
      view(jc, jk) = p;
    }
  }
};

#if 0
void scenario_5(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print) std::cout << "scenario 5: Adaptable layout & functor & subview (array, nproma, nlev);" << std::endl;

  Kokkos::View<double***, Layout, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view   = Kokkos::create_mirror_view_and_copy(space_t(), view);

  timer.reset();
  for (int jb = 0; jb < nblocks; ++jb) {
    LoopFunctor<double> loop(nproma, nlev, nblocks, jb, d_view);
    Kokkos::parallel_for(Kokkos::RangePolicy<>(0, nproma), loop);
  }

  Kokkos::fence();
  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}
#endif

int KOKKOS_FUNCTION get_indexes(int nproma) {

  return nproma + 1 ;
}

#if 0
void scenario_6(double* array, int nblocks, int nlev, int nproma, bool print = true) {
  if (print) std::cout << "scenario 6: Adaptable Layout &  Hierarchical parallelism" << std::endl;

  Kokkos::View<double***, Layout, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  using space_t     = Kokkos::DefaultExecutionSpace::memory_space;
  auto d_view       = Kokkos::create_mirror_view_and_copy(space_t(), view);

  using team_policy = Kokkos::TeamPolicy<>;
  using member_type = Kokkos::TeamPolicy<>::member_type;

  timer.reset();

  Kokkos::parallel_for(
      "blocks", team_policy(nblocks, Kokkos::AUTO), KOKKOS_LAMBDA(const member_type& teamMember) {
        const int jb = teamMember.league_rank();

        volatile int nproma_tmp = get_indexes(nproma -  1);  
        Kokkos::parallel_for(Kokkos::TeamThreadRange(teamMember, nproma_tmp), [&](const int jc) {
          // sequential over the levels
          for (int jk = 0; jk < nlev; ++jk) {
#if defined(DEMO_DEVICE)
            int p = jb * nlev * nproma + jk * nproma + jc;
#else 
                            int p = jc * nlev * nblocks + jk * nblocks + jb;
#endif

            d_view(jc, jk, jb) = p;
          }
        });
      });

  Kokkos::fence();

  if (print) printf("Time = %f ms\n\n", timer.seconds() * 1000);
  Kokkos::deep_copy(view, d_view);
  validate(array, nblocks, nlev, nproma);
}
#endif


void scenario_6b(double* array, int nblocks, int nlev, int nproma) {
  std::cout << "scenario 6b: Adaptable Layout &  Hierarchical parallelism" << std::endl;
  //exit(1);
  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);

  using team_policy = Kokkos::TeamPolicy<>;
  using member_type = Kokkos::TeamPolicy<>::member_type;

  timer.reset();

  Kokkos::parallel_for(
      "blocks", team_policy(nblocks, Kokkos::AUTO), KOKKOS_LAMBDA(const member_type& teamMember) {
        const int jb = teamMember.league_rank();
        Kokkos::parallel_for(Kokkos::TeamThreadRange(teamMember, nproma), [&](const int jc) {
          update_field(d_view, jb, all_levels, jc);
        });
      });

  Kokkos::fence();

  printf("Time = %f ms\n\n", timer.seconds() * 1000);
  //validate(array, nblocks, nlev, nproma);
}


#if 0
void scenario_7(double* array, int nblocks, int nlev, int nproma, bool print=true) {

    if(print)
        std::cout << "scenario 7: Default layout; view(array, nblocks, nlev, nproma); d_view(jb, jk, jc) ----- " << std::endl;

    Kokkos::View<double***, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nblocks, nlev, nproma);
 
    using space_t = Kokkos::DefaultExecutionSpace::memory_space;
    auto d_view = Kokkos::create_mirror_view_and_copy(space_t(), view);
    
    timer.reset();
    Kokkos::parallel_for("", Kokkos::RangePolicy<>(0, nblocks), KOKKOS_LAMBDA (const int jb) {
      //for (int jb = 0 ; jb < nblocks; ++jb) {
      for (int jc = 0 ; jc < nproma; ++jc) {
        for (int jk = 0; jk < nlev; ++jk) {
            int p = jb * nlev * nproma + jk * nproma + jc;
            check_bounds(jb,jk,jc, d_view.extent(0), d_view.extent(1), d_view.extent(2));
            d_view(jb, jk, jc) = p;
            //  printf("%f ", d_view(jb, jk, jc));      
          }
        }
      });

    Kokkos::fence();

    if(print)
        printf("Time = %f ms\n\n", timer.seconds() * 1000);
    Kokkos::deep_copy(view, d_view);
}
#endif

void scenario_7b(double* array, int nblocks, int nlev, int nproma) {
  std::cout << "scenario 7b: +ACC Default layout; CPUview(array, blocks, nlev, nproma); GPUview(nproma, nlev, nblocks), parallel blocks ----- " << std::endl;
#ifdef DEMO_DEVICE
  Kokkos::View<double***, Kokkos::MemoryUnmanaged> d_view(array, nproma, nlev, nblocks);
#else
  Kokkos::View<double***, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);
#endif
  timer.reset();
  Kokkos::parallel_for("", Kokkos::RangePolicy<>(0, nblocks), KOKKOS_LAMBDA (const int jb) {
      for (int jc = 0 ; jc < nproma; ++jc) {
#ifdef DEMO_DEVICE
        update_field(d_view, jc, all_levels, jb);
#else
        update_field(d_view, jb, all_levels, jc);
#endif        
      }
    });
  Kokkos::fence();
  printf("Time = %f ms\n\n", timer.seconds() * 1000);
}




void scenario_7bmacro(double* array, int nblocks, int nlev, int nproma) {

  std::cout << "scenario 7bmacro: +ACC CPU:LayoutLeft GPU:LayoutRight; CPUview(array, nblocks, nlev, nproma); GPUview(nproma, nlev, nblocks) para: asICON" << std::endl;

#if defined(DEMO_DEVICE)
  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> d_view(array, nproma, nlev, nblocks);
#else
  Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);
#endif
  timer.reset();
  outer_for("", 0, nblocks, outer_lambda(const int jb){
      inner_for("",0, nproma, inner_lambda(const int jc) {
#ifdef DEMO_DEVICE
          update_field(d_view, jc, all_levels, jb);
#else
          update_field(d_view, jb, all_levels, jc);
#endif
        });
    });
  Kokkos::fence();
  printf("Time = %f ms\n\n", timer.seconds() * 1000);
}

void scenario_7btemplate(double* array, int nblocks, int nlev, int nproma) {

    std::cout << "scenario 7btemplate: +ACC allways LayoutRight; view(array, nblocks, nlev, nproma); parallel: asICON ----- " << std::endl;

    Kokkos::View<double***, Kokkos::LayoutRight, Kokkos::MemoryUnmanaged> d_view(array, nblocks, nlev, nproma);

    timer.reset();
    outer_loop(nblocks, outer_lambda(const int jb){
        inner_loop(nproma, inner_lambda(const int jc) {
            update_field(d_view, jb, all_levels, jc);
          });
      });
    Kokkos::fence();
    printf("Time = %f ms\n\n", timer.seconds() * 1000);
}


void scenario_8(double* array, int nblocks, int nlev, int nproma) {
    std::cout << "scenario 8: +ACC always LayoutLeft view(array, nproma, nlev, nblocks) parallel: nproma ----- " << std::endl;

    Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> d_view(array, nproma, nlev, nblocks) ;
    timer.reset();
    for (int jb = 0; jb < nblocks; ++jb)
      Kokkos::parallel_for("", Kokkos::RangePolicy<>(0, nproma), KOKKOS_LAMBDA(const int jc) {
          update_field(d_view, jc, all_levels, jb);
        });
    Kokkos::fence();

    printf("Time = %f ms\n\n", timer.seconds() * 1000);
}


void scenario_8macro(double* array, int nblocks, int nlev, int nproma) {
    std::cout << "scenario 8macro: +ACC always LayoutLeft, view(array, nproma, nlev, nblocks); parallel: asICON ----- " << std::endl;
    using space_t = Kokkos::DefaultExecutionSpace::memory_space;
    Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::MemoryUnmanaged> d_view(array, nproma, nlev, nblocks);
    timer.reset();
    outer_for("", 0, nblocks, outer_lambda(const int jb) {
        inner_for("",0, nproma, inner_lambda(const int jc) {
            update_field(d_view, jc, all_levels, jb);
        });
      });
    Kokkos::fence();
    printf("Time = %f ms\n\n", timer.seconds() * 1000);
}

void scenario_8template(double* array, int nblocks, int nlev, int nproma) {
    std::cout << "scenario 8template: +ACC always LayoutLeft, view(array, nproma, nlev, nblocks); parallel: asICON ----- " << std::endl;
    using space_t = Kokkos::DefaultExecutionSpace::memory_space;
    Kokkos::View<double***, Kokkos::LayoutLeft> d_view(array, nproma, nlev, nblocks);
    timer.reset();
    outer_loop(nblocks, outer_lambda(const int jb) {
        inner_loop(nproma, inner_lambda(const int jc) {
            update_field(d_view, jc, all_levels, jb);
        });
      });
    Kokkos::fence();
    printf("Time = %f ms\n\n", timer.seconds() * 1000);
}




void openacc_calls(double* array, const double *ref, int nblocks, int nlev, int nproma, 
                      std::function<void(double*, int, int, int)> func){
#if defined(DEMO_DEVICE)
    #pragma acc enter data copyin(array[0:nblocks*nlev*nproma]) 
    {
      #pragma acc host_data use_device(array) 
      {
        func(array, nblocks, nlev, nproma);
      }
    }
    #pragma acc exit data copyout(array[0:nblocks*nlev*nproma])
#else
      func(array, nblocks, nlev, nproma);
#endif
      validate(array, ref, nblocks, nlev, nproma);
   
}

void init_array(double *array, int nproma, int nlev, int nblocks) {
  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  for (int jb=0; jb < nblocks; ++jb) {
    for (int jk=0; jk < nlev; ++jk) {
      for (int jc=0; jc < nproma; ++jc) {
        view(jc,jk,jb) = jc + jk*nproma + jb * nproma*nblocks;
        //printf("init: jc=%d, jk=%d, jb=%d, view=%f\n",jc,jk,jb,view(jc,jk,jb));
      }
    }
  }
}

void physop(double *array, int nproma, int nlev, int nblocks) {
  Kokkos::View<double***, Kokkos::LayoutLeft, Kokkos::HostSpace, Kokkos::MemoryUnmanaged> view(array, nproma, nlev, nblocks);

  for (int jb=0; jb < nblocks; ++jb) {
    for (int jc=0; jc < nproma; ++jc) {
      update_field(view, jc, all_levels, jb);
    }
  }
}

bool is_device_ptr(const void* ptr) {
  using ExecSpace = Kokkos::DefaultExecutionSpace;
#ifdef KOKKOS_ENABLE_CUDA
  constexpr bool exec_on_cuda = std::is_same<ExecSpace,Kokkos::Cuda>::value;
  if constexpr (exec_on_cuda) {
    cudaPointerAttributes att;
    auto err = cudaPointerGetAttributes (&att, ptr);
    assert(err == cudaSuccess);
    if (verbose_level > 2 ) {
      printf("is_device_ptr: device=%d\n",att.device);
      printf("is_device_ptr: memtype=%d\n",att.type );
      printf("is_device_ptr:   ptr=%p\n",ptr);
      printf("is_device_ptr: d_ptr=%p\n",att.devicePointer);
      printf("is_device_ptr: h_ptr=%p\n",att.hostPointer);
    }
    if (att.devicePointer && ptr == att.devicePointer ) {
      return true;
    } else {
      return false;
    }
  }
#endif
#ifdef KOKKOS_ENABLE_HIP
  constexpr bool exec_on_hip = std::is_same<ExecSpace,Kokkos::HIP>::value;
  if constexpr (exec_on_hip) {
    hipPointerAttribute_t att;
    auto err = hipPointerGetAttributes(&att, ptr);
    assert(err == hipSuccess);
    if (att.devicePointer && ptr == att.devicePointer ) {
      return true;
    } else {
      return false;
    }
  }
#endif

  return false;

}
    
void run_scenario_using_host_data(double *array, const double *array_in_ref, const double *array_out_ref,
                  int nblocks, int nlev, int nproma,
                  const std::function<void(double*, int, int, int)> &scenario) {
  const size_t nbytes = nproma*nlev*nblocks*sizeof(double);
  memcpy(array, array_in_ref, nbytes);
  openacc_calls(array, array_out_ref, nblocks, nlev, nproma, scenario);
}


void default_run_scenario(double *array, const double *array_in_ref, const double *array_out_ref,
                                    int nblocks, int nlev, int nproma,
                                    const std::function<void(double*, int, int, int)> &scenario) {
  // init:
  Kokkos::View<double*, Kokkos::MemoryUnmanaged> view(array, nblocks*nlev*nproma);
  Kokkos::View<const double*, Kokkos::MemoryUnmanaged> view_in_ref(array_in_ref, nblocks*nlev*nproma);
  Kokkos::deep_copy(view, view_in_ref);

  // run:
  scenario(view.data(), nblocks, nlev, nproma);

  // check:
  validate_max_error(array, array_out_ref, nblocks, nlev, nproma);
}

struct Domain {
  int ncells;
  int nlev;
  int nblocks;
  int nproma;
};

int get_int(const char* str, const int default_value) {
  const char* istr = std::getenv(str);
  if (not istr) {
    printf("get_int: key=%s, using default=%d\n",str, default_value);
    return default_value;
  }
  assert(istr);
  int64_t i64 = atoi(istr);
  assert(i64 >= INT_MIN and i64 <= INT_MAX);
  int i = i64;
  return i;
}

Domain get_domain() {
  int ncells = get_int("NCELLS", 0);
  int nproma  = get_int("NPROMA", 0);
  int nlev  = get_int("NLEV", 0);
  int nblocks = (ncells - 1) / nproma + 1;
  assert(nproma*nblocks == ncells);
  Domain dom;
  dom.ncells  = ncells;
  dom.nlev    = nlev;
  dom.nproma  = nproma;
  dom.nblocks = nblocks;
  return dom;
}

extern "C"
void f2c_get_domain(Domain *dom) {
  Domain td = get_domain();
  *dom = get_domain();
}


void default_run_all(double* field, double* ref_in, double* ref_out, int nblocks, int nlev, int nproma) {
  verbose_level=get_int("VERBOSE_LEVEL", 0);
  if (verbose_level > 0) printf("default_run_all: start\n");

  assert(is_device_ptr(field) == is_device_ptr(ref_in));
  assert(is_device_ptr(field) == is_device_ptr(ref_out));

  std::vector funs = {
    scenario_1_noprint,
    scenario_1b, scenario_1c, scenario_1cmacro,
    scenario_6b,
    scenario_7bmacro, scenario_7btemplate,
    scenario_8, scenario_8macro, scenario_8template
  };

  if (exec_on_device() and (nproma > 50000)) {
    printf("skip scenario_7b: GPU parallel over blocks way too slow\n\n");
  } else {
    funs.push_back(scenario_7b);
  }

  for (const auto &f : funs) {
    default_run_scenario(field, ref_in, ref_out,
                         nblocks, nlev, nproma,
                         f);
    printf("====\n");
  }
  if (verbose_level > 0) printf("default_run_all: end\n");
}

extern "C"
void f2c_default_run_all(double* field, double* ref_in, double* ref_out, int nblocks, int nlev, int nproma) {
  default_run_all(field, ref_in, ref_out, nblocks, nlev, nproma);
}

extern "C" void f2c_kokkos_initialize() {
  Kokkos::initialize();
}

extern "C" void f2c_kokkos_finalize() {
  Kokkos::finalize();
}

void cpp_run_tests() {
  const Domain dom = get_domain();
  int nlev    = dom.nlev;
  int nproma  = dom.nproma;
  int nblocks = dom.nblocks;

  verbose_level=get_int("VERBOSE_LEVEL", 0);

  std::cout << "nblocks=" << nblocks << ", nlev=" << nlev << ", nproma=" << nproma << std::endl;

  double array_out_ref[nblocks * nlev * nproma];
  double array_in_ref[nblocks * nlev * nproma];
  double array[nblocks * nlev * nproma];
  const size_t nbytes = nproma*nlev*nblocks*sizeof(double);
  {

    init_array(array_in_ref, nproma,nlev,nblocks);

    memcpy(array_out_ref, array_in_ref, nbytes);
    physop(array_out_ref, nproma,nlev,nblocks);

    memcpy(array, array_in_ref, nbytes);
    scenario_1(array, nblocks, nlev, nproma, false);

    //memcpy(array, array_in_ref, nbytes);
    //scenario_1(array, nblocks, nlev, nproma);
    //validate(array, array_out_ref, nblocks, nlev, nproma);
    run_scenario_using_host_data(array, array_in_ref, array_out_ref,
                                 nblocks, nlev, nproma,
                                 scenario_6b);
    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_1b = scenario_1b;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_1b);

    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_1c = scenario_1c;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_1c);

    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_1cmacro = scenario_1cmacro;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_1cmacro);

    if (exec_on_device() and (nproma > 50000)) {
      printf("skip scenario_7b: GPU parallel over blocks way too slow\n\n");
    } else {
      memcpy(array, array_in_ref, nbytes);
      std::function<void(double*, int, int, int)> s_7b = scenario_7b;
      openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_7b);
    }

    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_7bmacro = scenario_7bmacro;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_7bmacro);

    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_7btemplate = scenario_7btemplate;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_7btemplate);



    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_8 = scenario_8;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_8);

    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_8macro = scenario_8macro;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_8macro);

    memcpy(array, array_in_ref, nbytes);
    std::function<void(double*, int, int, int)> s_8template = scenario_8template;
    openacc_calls(array, array_out_ref, nblocks, nlev, nproma, s_8template);

    //    exit(0);
/*
    scenario_2(array, nblocks, nlev, nproma);
    scenario_2b(array, nblocks, nlev, nproma);
    scenario_3(array, nblocks, nlev, nproma);
    scenario_4(array, nblocks, nlev, nproma);
    scenario_4b(array, nblocks, nlev, nproma);
    scenario_5(array, nblocks, nlev, nproma);
    scenario_6(array, nblocks, nlev, nproma);

    scenario_7(array, nblocks, nlev, nproma);
    validate(array, array_out_ref, nblocks, nlev, nproma);
*/

  }


}
